//-----------------------------------------------------------------------------
// File: fgrp_Circuit.hh
//
// Purpose: Declaration of the Circuit related structures
//
// History: 03/10/98 - LGS - created.
//
// Copyright (c) 1999 ALGOS Group / INESC
//
//-----------------------------------------------------------------------------

#include "fcgrp_Circuit.hh"

Node::Node(int i)
  : fanins(D_GROW_ONLY), fanouts(D_GROW_ONLY)
{
  ID = i;
  type = INPUT;
  depth = -1;
  jc[0] = 0;
  jc[1] = 0;
  invert = 0;
  variable = NULL;
  fanins.resize(0);
  fanouts.resize(0);
}

Node::~Node()
{
  fanins.resize(0);
  fanouts.resize(0);
}

void Node::unconnect ()
{
  register int i;

  for (i = 0; i < fanins.size(); i++)
    fanins[i]->del_fanout(this);  
  for (i = 0; i < fanouts.size(); i++)
    fanouts[i]->del_fanin(this);  
}

void Node::del_fanout (NodePtr fanout)
{
  register int i;

  for (i = 0; i < fanouts.size(); i++)
    if (fanouts[i] == fanout) {
      fanouts[i] = fanouts[fanouts.size() - 1];
      fanouts.decr_size();
      return;
    }
  if (fanouts.size() == 0)
    type = OUTPUT;
}

void Node::del_fanin (NodePtr fanin)
{
  register int i;

  for (i = 0; i < fanins.size(); i++)
    if (fanins[i] == fanin) {
      fanins[i] = fanins[fanins.size() - 1];
      fanins.decr_size();
      return;
    }
  if (fanins.size() == 0)
    type = INPUT;
}

void Node::update_fanouts (int value, int count)
{
  register int i;
  register int fsize = fanouts.size();

  for (i = 0; i < fsize; i++)
    if (fanouts[i]->type != OUTPUT)
      fanouts[i]->jc[value] += count;
}

int Node::is_justified ()
{
  int value = variable->value();
  int just = 0;

  assert (value != UNKNOWN);

  if (type == INPUT)
    just = 1;
  else if (type == OUTPUT)
    just = fanins[0]->is_justified();
  else if (type == AND || type == OR || type == BUFF)
    just = (jc[value] <= 0);
  else if (type == NAND || type == NOR || type == NOT)
    just = (jc[1 - value] <= 0);
  else if (type == XOR || type == XNOR)
    just = ((jc[0] + jc[1]) == 2);
  else if (type == ITE) {
    if (fanins[1]->variable->value() == value &&
        fanins[2]->variable->value() == value)
      just = 1;
    else if (fanins[0]->variable->value() == 1 &&
             fanins[1]->variable->value() == value)
      just = 1;
    else if (fanins[0]->variable->value() == 0 &&
             fanins[2]->variable->value() == value)
      just = 1;
    else
      just = 0;
  }

  return just;
}

Circuit::Circuit(ClauseDatabase &clDB)
  : _clDB(clDB), nodes(D_GROW_ONLY), _Jf_data(D_GROW_ONLY)
{
}

Circuit::~Circuit()
{ 
  nodes.resize(0);
  _Jf_data.resize(0);
  jFrontier.erase();
}

void Circuit::update_fanouts (int varID, int value, int count)
{ 
  nodes[varID]->update_fanouts (value, count);
}

void Circuit::add_PO (NodePtr node)
{
  NodePtr nw;
  char buffer[BUFSIZ];

  sprintf (buffer, "OUT_%s", node->name);
  nw = add_node (strsav(buffer)); 
  nw->type = OUTPUT;
  nw->add_fanin (node);
  node->add_fanout (nw);
  POs[POs.incr_size()] = nw;
}

Node *Circuit::add_node (char *name)
{
  Node *node;
  VariablePtr var;
  int varID;
  int size;

  size = nodes.size();

  var = _clDB.add_variable(name);
  varID = var->ID();

  if (varID >= size) {
    nodes.incr_size(varID - size + 1);
    _Jf_data.incr_size(varID - size + 1);
  }
  node = new Node(varID);
  nodes[varID] = node;
  node->variable = var;
  node->name = var->name();
  _Jf_data[varID] = NULL;

  return node;
}

void Circuit::dump()
{
  register int i, j;
  static char *types[] = { "INPUT", "AND", "NAND", "OR", "NOR", "XOR", "XNOR",
                           "BUFF", "NOT", "ITE", "OUTPUT", "BAD", "VIRTUAL" };
  printf ("PIs:");
  for (i = 0; i < PIs.size(); i++)
    printf (" %s", PIs[i]->name);
  printf ("\n");
  printf ("POs:");
  for (i = 0; i < POs.size(); i++)
    printf (" %s", POs[i]->name);
  printf ("\n");
  printf ("NODES:\n");
  for (i = 1; i < nodes.size(); i++) {
    printf ("%s %s[%d]\n", types[nodes[i]->type], nodes[i]->name,
                           nodes[i]->depth); 
    printf ("  FANINS: ");
    for (j = 0; j < nodes[i]->fanins.size(); j++)
      printf (" %s", nodes[i]->fanins[j]->name);
    printf ("\n");
    printf ("  FANOUTS: ");
    for (j = 0; j < nodes[i]->fanouts.size(); j++)
      printf (" %s", nodes[i]->fanouts[j]->name);
    printf ("\n");

  }
  printf ("\n");
}

void Circuit::add_gate (int type, NodePtr output, NodePtr input1,
                        NodePtr input2)
{
  NodePtr inputs[2];

  inputs[0] = input1;
  inputs[1] = input2;
  add_gate (type, output, inputs, 2);
}

void Circuit::add_gate (int type, NodePtr output, NodePtr input)
{
  NodePtr inputs[1];

  inputs[0] = input;
  add_gate (type, output, inputs, 1);
}

void Circuit::add_gate (int type, NodePtr output, NodePtr *inputs,
                        int Ninputs)
{
  register int i;
  VariablePtr v_output, *v_inputs;
  ClausePtr clause;

  v_inputs = new VariablePtr[Ninputs];

  for (i = 0; i < Ninputs; i++) {
    output->add_fanin (inputs[i]);
    inputs[i]->add_fanout (output);
    v_inputs[i] = inputs[i]->variable;
  }
  v_output = output->variable;

  switch(type) {
    case AND:
      output->jc[0] = 1;
      output->jc[1] = Ninputs;
      output->type = AND;
      output->invert = 0;
      output->ctrlvalue = 0;
      output->ctrlfanin = 0;
      clause = _clDB.add_clause();
      _clDB.add_literal (clause, v_output, FALSE);
      for (i = 0; i < Ninputs; i++)
        _clDB.add_literal (clause, v_inputs[i], TRUE);
      for (i = 0; i < Ninputs; i++) {
        clause = _clDB.add_clause();
        _clDB.add_literal (clause, v_inputs[i], FALSE);
        _clDB.add_literal (clause, v_output, TRUE);
      }
      break;
    case NAND:
      output->jc[0] = 1;
      output->jc[1] = Ninputs;
      output->type = NAND;
      output->invert = 1;
      output->ctrlvalue = 1;
      output->ctrlfanin = 0;
      clause = _clDB.add_clause();
      _clDB.add_literal (clause, v_output, TRUE);
      for (i = 0; i < Ninputs; i++)
        _clDB.add_literal (clause, v_inputs[i], TRUE);
      for (i = 0; i < Ninputs; i++) {
        clause = _clDB.add_clause();
        _clDB.add_literal (clause, v_inputs[i], FALSE);
        _clDB.add_literal (clause, v_output, FALSE);
      }
      break;
    case OR:
      output->jc[0] = Ninputs;
      output->jc[1] = 1;
      output->type = OR;
      output->invert = 0;
      output->ctrlvalue = 1;
      output->ctrlfanin = 1;
      clause = _clDB.add_clause();
      _clDB.add_literal (clause, v_output, TRUE);
      for (i = 0; i < Ninputs; i++)
        _clDB.add_literal (clause, v_inputs[i], FALSE);
      for (i = 0; i < Ninputs; i++) {
        clause = _clDB.add_clause();
        _clDB.add_literal (clause, v_inputs[i], TRUE);
        _clDB.add_literal (clause, v_output, FALSE);
      }
      break;
    case NOR:
      output->jc[0] = Ninputs;
      output->jc[1] = 1;
      output->type = NOR;
      output->invert = 1;
      output->ctrlvalue = 0;
      output->ctrlfanin = 1;
      clause = _clDB.add_clause();
      _clDB.add_literal (clause, v_output, FALSE);
      for (i = 0; i < Ninputs; i++)
        _clDB.add_literal (clause, v_inputs[i], FALSE);
      for (i = 0; i < Ninputs; i++) {
        clause = _clDB.add_clause();
        _clDB.add_literal (clause, v_inputs[i], TRUE);
        _clDB.add_literal (clause, v_output, TRUE);
      }
      break;
    case BUFF:
      output->jc[0] = 1;
      output->jc[1] = 1;
      output->type = BUFF;
      output->invert = 0;
      output->ctrlvalue = 4;
      output->ctrlfanin = 4;
      clause = _clDB.add_clause();
      _clDB.add_literal (clause, v_output, FALSE);
      _clDB.add_literal (clause, v_inputs[0], TRUE);
      clause = _clDB.add_clause();
      _clDB.add_literal (clause, v_output, TRUE);
      _clDB.add_literal (clause, v_inputs[0], FALSE);
      break;
    case NOT:
      output->jc[0] = 1;
      output->jc[1] = 1;
      output->type = NOT;
      output->invert = 1;
      output->ctrlvalue = 4;
      output->ctrlfanin = 4;
      clause = _clDB.add_clause();
      _clDB.add_literal (clause, v_output, TRUE);
      _clDB.add_literal (clause, v_inputs[0], TRUE);
      clause = _clDB.add_clause();
      _clDB.add_literal (clause, v_output, FALSE);
      _clDB.add_literal (clause, v_inputs[0], FALSE);
      break;
    case XOR:
      if (Ninputs > 2)
        printf ("error: XOR gates with more than 2 inputs are not allowed\n");
      output->jc[0] = 2;
      output->jc[1] = 2;
      output->type = XOR;
      output->invert = -2;
      output->ctrlvalue = 4;
      output->ctrlfanin = 4;
      clause = _clDB.add_clause();
      _clDB.add_literal(clause, v_output, TRUE);
      _clDB.add_literal(clause, v_inputs[0], FALSE);
      _clDB.add_literal(clause, v_inputs[1], FALSE);
      clause = _clDB.add_clause();
      _clDB.add_literal(clause, v_output, TRUE);
      _clDB.add_literal(clause, v_inputs[0], TRUE);
      _clDB.add_literal(clause, v_inputs[1], TRUE);
      clause = _clDB.add_clause();
      _clDB.add_literal(clause, v_output, FALSE);
      _clDB.add_literal(clause, v_inputs[0], FALSE);
      _clDB.add_literal(clause, v_inputs[1], TRUE);
      clause = _clDB.add_clause();
      _clDB.add_literal(clause, v_output, FALSE);
      _clDB.add_literal(clause, v_inputs[0], TRUE);
      _clDB.add_literal(clause, v_inputs[1], FALSE);
      break;
    case XNOR:
      if (Ninputs > 2)
        printf ("error: XNOR gates with more than 2 inputs are not allowed\n");
      output->jc[0] = 2;
      output->jc[1] = 2;
      output->type = XNOR;
      output->invert = -3;
      output->ctrlvalue = 4;
      output->ctrlfanin = 4;
      clause = _clDB.add_clause();
      _clDB.add_literal(clause, v_output, FALSE);
      _clDB.add_literal(clause, v_inputs[0], FALSE);
      _clDB.add_literal(clause, v_inputs[1], FALSE);
      clause = _clDB.add_clause();
      _clDB.add_literal(clause, v_output, FALSE);
      _clDB.add_literal(clause, v_inputs[0], TRUE);
      _clDB.add_literal(clause, v_inputs[1], TRUE);
      clause = _clDB.add_clause();
      _clDB.add_literal(clause, v_output, TRUE);
      _clDB.add_literal(clause, v_inputs[0], FALSE);
      _clDB.add_literal(clause, v_inputs[1], TRUE);
      clause = _clDB.add_clause();
      _clDB.add_literal(clause, v_output, TRUE);
      _clDB.add_literal(clause, v_inputs[0], TRUE);
      _clDB.add_literal(clause, v_inputs[1], FALSE);
      break;
    case ITE:
      output->type = ITE;
      output->invert = -3;
      output->ctrlvalue = 4;
      output->ctrlfanin = 4;
      clause = _clDB.add_clause();
      _clDB.add_literal(clause, v_inputs[0], TRUE);
      _clDB.add_literal(clause, v_output, TRUE);
      _clDB.add_literal(clause, v_inputs[1], FALSE);
      clause = _clDB.add_clause();
      _clDB.add_literal(clause, v_inputs[0], TRUE);
      _clDB.add_literal(clause, v_output, FALSE);
      _clDB.add_literal(clause, v_inputs[1], TRUE);
      clause = _clDB.add_clause();
      _clDB.add_literal(clause, v_inputs[0], FALSE);
      _clDB.add_literal(clause, v_output, TRUE);
      _clDB.add_literal(clause, v_inputs[2], FALSE);
      clause = _clDB.add_clause();
      _clDB.add_literal(clause, v_inputs[0], FALSE);
      _clDB.add_literal(clause, v_output, FALSE);
      _clDB.add_literal(clause, v_inputs[2], TRUE);
      break;
    default:
      break;
  }
  delete v_inputs;
}

void Circuit::jf_add (int ID)
{
  if (_Jf_data[ID] == NULL && nodes[ID]->type != INPUT &&
      nodes[ID]->type != OUTPUT)
    _Jf_data[ID] = jFrontier.append(ID);
}

void Circuit::jf_del (int ID)
{
  if (_Jf_data[ID] != NULL) {
    jFrontier.remove (_Jf_data[ID]);
    _Jf_data[ID] = NULL;
  }
}

int Circuit::jf_get (int ID)
{
  if (_Jf_data[ID] == NULL)
    return 0;
  else
    return 1;
}

VariablePtr Circuit::jf_dequeue ()
{
  if (jFrontier.last() != NULL)
    return nodes[jFrontier.last()->data()]->variable;
 else
    return NULL;
}

void Circuit::checkout_jf_on_fanouts (int varID)
{
  register int i;
  DNodeArray &fanouts = nodes[varID]->fanouts;
  register int size;

  size = fanouts.size();

  for (i = 0; i < size; i++) {
    Node *node = fanouts[i];
    if (_Jf_data[node->ID] != NULL && node->is_justified() == 1)
      jf_del (node->ID);
  }
}

void Circuit::checkout_jf_on_var (int varID)
{
  if (_Jf_data[varID] != NULL && nodes[varID]->is_justified() == 1)
    jf_del (varID);
}

void Circuit::checkin_jf_on_fanouts (int varID)
{
  register int i;
  DNodeArray &fanouts = nodes[varID]->fanouts;
  register int size = fanouts.size();
 
  for (i = 0; i < size; i++) {
    Node *node = fanouts[i];
    if (node->value() != UNKNOWN && _Jf_data[node->ID] == NULL &&
        node->is_justified() == 0)
      jf_add (node->ID);
  }    
}

void Circuit::calc_depth ()
{
  register int i;
  int max, d;
  List <NodePtr> queue;
  NodePtr node;
  int counters[nodes.size()];

  for (i = 1; i < nodes.size(); i++)
     counters[i] = nodes[i]->fanins.size(); 
  for (i = 0; i < PIs.size(); i++) {
    queue.append (PIs[i]);
    PIs[i]->depth = 0;
  }
  while (node = queue.pop()) {
    max = -1;
    for (i = 0; i < node->fanins.size(); i++) {
      d = node->fanins[i]->depth;
      if (d > max)
        max = d;
    }
    node->depth = max+1;
    for (i = 0; i < node->fanouts.size(); i++) {
      counters[node->fanouts[i]->ID]--;
      if (counters[node->fanouts[i]->ID] == 0)
        queue.append (node->fanouts[i]);
    }
  }
}

void Circuit::checkin_jf_on_var (int varID)
{
  Node *node = nodes[varID];

  if (node->value() != UNKNOWN && _Jf_data[varID] == NULL &&
      node->is_justified() == 0)
    jf_add (varID);
}

int Circuit::load_netlist_iscas85 (char *filename)
{
  FILE *in;
  register int c;
  register int i;
  register int flag;
  int nargs;
  int sign;
  char buffer[LARGE_BUFSIZ];
  char *args[300], *var_name;
  int isio, iscomment;
  SymbolTable <NodePtr> node_cache (1024);
  ClausePtr clause;
  NodePtr input[1000], output, tmpvar;
  int type;

  if ((in = fopen (filename, "r")) == NULL)
    return 0;

  c = 10;
  while(c != EOF) {

    i = 0;
    iscomment = 0;
    isio = 1;
    nargs = 0;
    args[nargs++] = &buffer[i];
    buffer[i] = '\0';
    flag = 0;

    for (c = getc(in); c != '\n' && c != EOF; c = getc(in)) {
      if (c == '#') {
        iscomment = 1;
        continue;
      } else if (iscomment == 1 && c != '@') { 
          for (c = getc(in); c != '\n' && c != EOF; c = getc(in));
          iscomment = 0;
          break;
      } else if (c == '@' && iscomment == 1) {
        buffer[i++] = c;
        buffer[i++] = '\0';
        args[nargs++] = &buffer[i];
        iscomment = 0;
        continue;
      } else if (c == '=')
        isio = 0;
      if (c != '=' && c != '(' && c != ',' && c != ')' &&
          c != ' ' && c != '\t' ) {
        buffer[i++] = c;
        flag = 1;
      }
      else {
        if (flag == 1 && c != '\t' && c != ' ') {
          buffer[i++] = '\0';
          args[nargs++] = &buffer[i];
          flag = 0;
        }
        else if (flag == 1 && c == ' ') {
          buffer[i++] = '\0';
          args[nargs++] = &buffer[i];
          flag = 0;
        }
      }
    }
    buffer[i++] = '\0';

    if (args[0][0] == '@') {
      
      clause = add_clause();
      for (i = 1; i < nargs; i++) {
        sign = FALSE;
        if (args[i][0] == '-')
          sign = TRUE; 
        else if (args[i][0] != '+')
          args[i]--;
        args[i]++;
        if ((output = node_cache.lookup (args[i])) == NULL) {
          var_name = strsav(args[i]);
          node_cache.insert (output = add_node(var_name), var_name);
        }
        add_literal (clause, output, sign);
      }
    }
    else if (buffer[0] != '\0') {
      if (isio == 1) {	/* INPUT */
        if (args[0][0] == 'I' || args[0][0] == 'i') {
          if ((tmpvar = node_cache.lookup (args[1])) == NULL) {
            var_name = strsav(args[1]);
            node_cache.insert (tmpvar = add_node(var_name), var_name);
            add_PI (tmpvar);
          }
          else
            add_PI (tmpvar);
        }
        else {		/* OUTPUT */
          if ((tmpvar = node_cache.lookup (args[1])) == NULL) {
            var_name = strsav(args[1]);
            node_cache.insert (tmpvar = add_node(var_name), var_name);
            add_PO (tmpvar);
          }
          else
            add_PO (tmpvar);
        }
      }
      else {
        type = -1;
        switch (args[1][2]) {
          case 'd':
          case 'D': 	/* AND */
            type = AND;
            break;
          case 'n':
          case 'N':	/* NAND */
            type = NAND;
            break; 
          case '\0':	/* OR */
            type = OR;
            break;
          case 't':
          case 'T':	/* NOT */
            type = NOT;
            break;
          case 'f':
          case 'F':	/* BUFF */
            type = BUFF; 
            break;
          case 'o':
          case 'O':	/* XNOR */
            type = XNOR;
            break;
          case 'r':
          case 'R':
            if (args[1][0] == 'n' || args[1][0] == 'N') {	/* NOR */
              type = NOR;
            }
	    else if (args[1][0] == 'x' || args[1][0] == 'X') {	/* XOR */
              type = XOR;
            }
            break;
          case 'e':
          case 'E':
            type = ITE;
            break;
          default: 
            type = -1;
            break;
        }
        if (type != -1) {
          nargs-=3;
          if ((output = node_cache.lookup (args[0])) == NULL) {
            var_name = strsav(args[0]);
            node_cache.insert (output = add_node(var_name), var_name);
          }
          for (i = 0; i < nargs; i++) {
            if ((input[i] = node_cache.lookup (args[i+2])) == NULL) {
              var_name = strsav(args[i+2]);
              node_cache.insert (input[i] = add_node (var_name), var_name);
            }
          }
          add_gate (type, output, input, nargs);
        }
      }
    }
  }
  fclose(in);
  node_cache.clear();

  calc_depth();

  return 1;
} 

void Circuit::reset()
{
  register int i;
  register int type;
  
  // reset jc counters to their original values
  for (i = 1; i < nodes.size(); i++) {
    type = nodes[i]->type;
    if (type == INPUT || type == OUTPUT)
      continue;
    else if (type == AND || type == NAND) {
      nodes[i]->jc[0] = 1;
      nodes[i]->jc[1] = nodes[i]->fanins.size();
    }
    else if (type == OR || type == NOR) {
      nodes[i]->jc[0] = nodes[i]->fanins.size();
      nodes[i]->jc[1] = 1;
    }
    else if (type == BUFF || type == NOT) {
      nodes[i]->jc[0] = 1;
      nodes[i]->jc[1] = 1;
    }
    else if (type == XOR || type == XNOR) {
      nodes[i]->jc[0] = 2;
      nodes[i]->jc[1] = 2;
    }
    else if (type == ITE) {
      nodes[i]->jc[0] = 3;
      nodes[i]->jc[1] = 3;
    }
  } 

  // empty the j-frontier
  for (i = 0; i < _Jf_data.size(); i++)
    _Jf_data[i] = NULL;
  jFrontier.erase();

}

void Circuit::jf_dump()
{
  ListItem<int> *item;

  printf ("JF: ");
  for (item = jFrontier.first(); item != NULL; item = item->next())
    printf ("%s %d", nodes[item->data()]->variable->name(),
            nodes[item->data()]->is_justified());
  printf ("\n");
}

int Circuit::jf_size()
{
  return jFrontier.size(); 
}

VariablePtr Circuit::simple_backtrace (int varID, int *value)
{
  register int i;
  NodePtr nextGoal = nodes[varID], bestGoal;
  register int d;
  int idx;

  while (nextGoal->fanins.size() != 0) {
    d = 10000;
    for (i = 0; i < nextGoal->fanins.size(); i++)
      if (nextGoal->fanins[i]->value() == UNKNOWN &&
          nextGoal->fanins[i]->depth < d) {
        d = nextGoal->fanins[i]->depth;
        bestGoal = nextGoal->fanins[i];
        idx = i;
      }
    if (d == 10000) {
      printf ("BIG error in simple backtrace\n");
      exit(-1);
    }
    if (nextGoal->type == NAND || nextGoal->type ==NOR ||
        nextGoal->type == NOT)
      *value = 1 - *value;
    else if (nextGoal->type == XOR) {
       if (nextGoal->fanins[1 - idx]->value() != UNKNOWN)
         *value = *value ^ nextGoal->fanins[1 - idx]->value();
       else
         *value = 0;
    }
    else if (nextGoal->type == XNOR) {
       if (nextGoal->fanins[1 - idx]->value() != UNKNOWN)
         *value = 1 - (*value ^ nextGoal->fanins[1 - idx]->value());
       else
         *value = 0;
    }
    nextGoal = bestGoal;
  }
  return nextGoal->variable;
}

VariablePtr Circuit::var_imp_clauses (int *value)
{
  register ClausePtr cl = NULL;

  for_each(pcl, _reg_clauses, ClausePtr) {
    cl = (ClausePtr) pcl->data();

    assert (cl->is_unsatisfied());

    if (!(cl->is_satisfied())) {
      DLiteralArray &pos_literals = cl->pos_literals();
      DLiteralArray &neg_literals = cl->neg_literals();

      for (register int i = 0; i < neg_literals.size(); i++)
        if (neg_literals[i]->variable()->value() == UNKNOWN) {
          *value = FALSE;
          return neg_literals[i]->variable();
        }

      for (register int i = 0; i < pos_literals.size(); i++)
        if (pos_literals[i]->variable()->value() == UNKNOWN) {
          *value = TRUE;
          return pos_literals[i]->variable();
        }

      printf ("severe error in var_imp_clauses\n");
      exit(-1);
    }
  }
  return NULL;
}
